#include "AdvRivals.h"

#include "CvGameCoreDLL.h"
#include "CvGlobals.h"
#include "CvInitCore.h"
#include "CvDefines.h"
#include "CvGameCoreDLL.h"

#include <climits>
#include <cmath>
#include <algorithm>
#include <vector>
#include <string>
#include <sstream>
using std::vector;
using std::wstring;
using std::wostringstream;
using std::ostringstream;


void AdvRivals::setInitialItems(GameSpeedTypes eSpeed) {

    reset();
    report = "AdvRivals game setup\n";
    this->eSpeed = eSpeed;
    speed = &GC.getGameSpeedInfo(eSpeed);
    double growthFactor = speed->getGrowthPercent() / 100.0;
    report += CvString::format("Recognized speed as %f\n", growthFactor);
    gatherCivs();
    nCivs = civs.size();
    int foundingPrice = scaleToGrowth(getFoundingPrice());
    report += CvString::format("Recognized price of a city as %d\n", foundingPrice);
    MajorCiv::setFoundingPrice(foundingPrice);
    assignPoints();
    sortByPoints();
    for(int i = 0; i < nCivs; i++)
        report += CvString::format("Points for id %d: %d\n",
                i, civs[i]->startPoints_configured());
    randPoints = GC.getDefineINT("ADVRI_RANDOMIZE-POINTS") > 0;
    if(randPoints) {
        randomizePoints();
        report += CvString::format("Points after randomization:\n");
        for(int i = 0; i < nCivs; i++) {
            civs[i]->assignStartPoints();
            report += CvString::format("id %d: %d\n", i, civs[i]->startPoints_actual());
        }
        sortByPoints(); // randomization can disrupt order by breaking ties
    }
    else for(int i = 0; i < nCivs; i++)
        civs[i]->assignStartPoints();
    rearrangeStartingPlots();
    updatePointsDisplayString();
    setResearchPercent();
    // log report about game setup
    gDLL->logMsg("debug.log", report, true, true);
}


int AdvRivals::computeStartTurn(int unmoddedStartTurn) {

    int r;
    EraTypes era = GC.getGameINLINE().getStartEra();
    if(era > 0) { // starting in Classical or later era
        int laterEraOffset = GC.getDefineINT("ADVRI_LATER-ERA-START-TURN-OFFSET");
        r = unmoddedStartTurn + scaleToGrowth(laterEraOffset);
    }
    else // starting in Ancient era
        r = unmoddedStartTurn + round((maxStartPoints() + meanStartPoints()) * 0.01);
        /* example: If the maximum of the start points is 1200 and the mean 670,
           then the game starts on turn 19, i.e. 3240 B.C. (normal game speed).
           (no need to call 'scaleToGrowth' here b/c the start points are already
           adjusted to the game speed setting.) */
    setBarbarianStartTurn(r);
    return r;
}


/* Decides randomly and based on assigned start points whether an AI civ
   should be allowed to build a world wonder. */
bool AdvRivals::oracleRefuses(int civId) {

    if(civId < 0 || civId >= nCivs || civs[civId]->isHuman())
        return false;
    MajorCiv* mciv = civs[civId];
    FAssert(mciv->id() == civId);
    if(mciv->era() > 0) // only Ancient era restricted
        return false;
    double prPermit = GC.getDefineINT("ADVRI_AI_ANCIENT-WONDER-BUILD_PERCENT") / 100.0;
  // prPermit = (rank(civId) + 1.0) / (1.7 * nCivs); // previous approach; might want to revisit
    return !bernoulliSuccess(prPermit);
}


wstring* AdvRivals::forSettingsScreen() {

    return pointsDisplayString;
}


int AdvRivals::barbarianStartTurn() {

    return m_barbarianStartTurn;
}


int AdvRivals::researchPercent() {

    return m_researchPercent;
}



/* no need to remind the player about forced peace when he/she only has
   a few free Warriors and Archers while the AI civs get Advanced Starts. */
bool AdvRivals::everShowPeaceReminder() {

	return GC.getGameINLINE().getStartEra() >= 2 ||
		   human->startPoints_actual() > 0;
}


void AdvRivals::write(FDataStreamBase* pStream) {

    pStream->Write(randPoints);
    pStream->Write(prefSecluded);
    pStream->Write(nCivs);
    for(int i = 0; i < nCivs; i++) {
        allPts[2 * i] = civs[i]->startPoints_configured();
        allPts[2 * i + 1] = civs[i]->startPoints_actual();
    }
    pStream->Write(2 * MAX_PLAYERS, allPts);
}


void AdvRivals::read(FDataStreamBase* pStream) {

    reset();
    pStream->Read(&randPoints);
    pStream->Read(&prefSecluded);
    pStream->Read(&nCivs);
    pStream->Read(2 * MAX_PLAYERS, allPts);
    // done loading from stream; now recreate objects
    report = "Loading advri data from savegame\n";
    eSpeed = GC.getGameINLINE().getGameSpeedType();
    speed = &GC.getGameSpeedInfo(eSpeed);
    bool humanLoaded = false;
    for(int i = 0; i < nCivs; i++) {
        CvPlayer* p = &GET_PLAYER((PlayerTypes)i);
        FAssert(p != NULL);
        report += CvString::format("Loaded player id %d to starting civs\n", i);
        MajorCiv* mciv = new MajorCiv(*p);
        mciv->setStartPoints_configured(allPts[2 * i]);
        mciv->setStartPoints_actual(allPts[2 * i + 1]);
        report += CvString::format("Start points: %d (configured), %d (actual)\n",
                mciv->startPoints_configured(), mciv->startPoints_actual());
        if(p->isHuman()) {
            human = mciv;
            report += CvString::format("Human player: id %d civs\n", i);
            humanLoaded = true;
        }
        civs.push_back(mciv);
    }
    FAssert(humanLoaded);
    report += CvString::format("Civ ids sorted by points:\n");
    sortByPoints();
    for(int i = 0; i < nCivs; i++)
        report += CvString::format("%d ", i);
    report += CvString::format("\n");
    updatePointsDisplayString();
    setResearchPercent();
    setBarbarianStartTurn(GC.getGameINLINE().getStartTurn());
    // log report about loading data from savegame
    gDLL->logMsg("debug.log", report, true, true);
}


AdvRivals::AdvRivals() : speed(NULL), pointsDisplayString(NULL), eSpeed(NO_GAMESPEED),
        m_researchPercent(-1), m_barbarianStartTurn(-1) {

    allPts = new int [2*MAX_PLAYERS];
}


AdvRivals::~AdvRivals() {

    for(int i = 0; i < nCivs; i++)
        delete civs[i];
    SAFE_DELETE_ARRAY(allPts);
    if(pointsDisplayString != NULL)
        delete pointsDisplayString;

}


void AdvRivals::gatherCivs() {

    report += CvString::format("MAX_PLAYERS: %d\n", MAX_PLAYERS);
    for(int i = 0; i < MAX_PLAYERS; i++) {
        // PlayerTypes are just player ids; the only type-like value is "NO_PLAYER"
        if((PlayerTypes)i == NO_PLAYER) {
            report += CvString::format("Removed non-player (index %d)\n", i);
            continue;
        }
        CvPlayer* p = &GET_PLAYER((PlayerTypes)i);
        if(p == NULL) {
            report += CvString::format("Player id %d is NULL\n", i);
            continue;
        }
        if(p->isBarbarian()) {
            report += CvString::format("Player id %d is Barbarian\n", i);
            continue;
        }
        if(!p->isAlive()) {
            report += CvString::format("Player id %d is not alive\n", i);
            continue;
        }
        if(p->isMinorCiv()) {
            report += CvString::format("Player id %d is a MinorCiv\n", i);
            continue;
        }
        FAssert(p->getID() == i);
        MajorCiv* mciv = new MajorCiv(*p);
        civs.push_back(mciv);
        report += CvString::format("Added player id %d to starting civs\n", i);
        if(p->isHuman()) {
            human = mciv;
            report += CvString::format("Human id: %d\n", i);
        }
    }
}


void AdvRivals::assignPoints() {

    report += CvString::format("Assigning points to %d known civs\n", nCivs);
    if(nCivs < 1)
        return;
    int playerPoints = scaleToGrowth(GC.getDefineINT("ADVRI_PLAYER-POINTS"));
    human->setStartPoints_configured(playerPoints);
    report += CvString::format("Player points set to %d\n", playerPoints);
    if(nCivs < 2 && playerPoints <= 0)
        return;
    if(GC.getDefineINT("ADVRI_POINTS_FROM_XML"))
        assignAIPoints_xml();
    else assignAIPoints_std();
}


void AdvRivals::assignAIPoints_std() {

    int pittancePercent = GC.getDefineINT("ADVRI_PITTANCE_PERCENT");
    CvInitCore& ic = GC.getInitCore();
    int pointsInput = ic.getNumAdvancedStartPoints();
    if(pointsInput <= 0) {
        report += "No points param from game options; using default\n";
        /* if player didn't tick "Advanced Start", then resetAdvancedStartPoints
           may not have been called. do it here. */
        ic.resetAdvancedStartPoints();
        pointsInput = ic.getNumAdvancedStartPoints();
    }
    else report += CvString::format("Game options have set points to %d\n", pointsInput);
    int pittance = round(pittancePercent * pointsInput / 100.0);
    pointsInput = scaleToGrowth(pointsInput);
    pittance = scaleToGrowth(pittance);
    report += CvString::format("Input points adjusted to game speed: %d\n", pointsInput);
    report += CvString::format("Pittance points adjusted to game speed: %d\n", pittance);
    /* TODO assuming that the human player is id=0 will cause problems when
            adding multiplayer support, if not earlier. should probably put
            all the AIs in one array in 'assignPoints' and pass that to
            this function and the _xml version */
    // rule 1: if there is a "middle" AI, that AI receives the mean of the weakest and toughest
    if(nCivs % 2 == 0)
        civs[nCivs / 2]->setStartPoints_configured(round((pointsInput + pittance) / 2.0));
    if(nCivs <= 4) {
        // rule 2: toughest civ receives exactly the points input (precedes over rule 3)
        civs[nCivs - 1]->setStartPoints_configured(pointsInput);
        // rule 3: weakest civ receives exactly the pittance
        if(nCivs > 2)
            civs[1]->setStartPoints_configured(pittance);
        return;
    }
    /* from 5 civs (4 AI) upwards, rule 4: form pairs of AIs that receive equal points
       and let points increase linearly from pair to pair, starting at the pittance
       and ending with the points input. */
    int nPairs = (nCivs - 1) / 2;
    double step = (pointsInput - pittance) / (nPairs - 1.0);
    double pts = pittance;
    for(int i = 1; i < nCivs - 1; i += 2) {
        if(nCivs % 2 == 0 && i == nCivs / 2)
            i++;
        civs[i]->setStartPoints_configured(round(pts));
        if(nCivs % 2 == 0 && i + 1 == nCivs / 2)
            i++;
        civs[i + 1]->setStartPoints_configured(
                civs[i]->startPoints_configured());
        pts += step;
    }
}


void AdvRivals::assignAIPoints_xml() {

    int rivalIndex = 1;
    for(int i = 0; i < nCivs; i++) {
        if(civs[i]->equals(human))
            continue;
        ostringstream s;
        s << "ADVRI_RIVAL" << rivalIndex;
        CvString msg("Looking up XML node ");
        msg += s.str() + "\n";
        report += msg;
        int pts = GC.getDefineINT(s.str().c_str());
        pts = scaleToGrowth(pts);
        civs[i]->setStartPoints_configured(pts);
        rivalIndex++;
    }
}


int AdvRivals::scaleToGrowth(int x) {

    return round(x * speed->getGrowthPercent() / 100.0);
}


void AdvRivals::randomizePoints() {

    if(nCivs <= 4)
        return;
    // TODO algorithm assumes player id=0; see comment in 'assignAIPoints'.
    // special pair [2], [n-2] (a sort of wrap-around)
    bounce(nCivs - 2, 2);
    if(nCivs <= 6)
        return;
    for(int i = 3; i <= nCivs - 3; i += 2) {
        if(nCivs % 2 == 0 && i == nCivs / 2)
            i++;
        if(nCivs % 2 == 0 && i + 1 == nCivs / 2) {
            bounce(i, i + 2);
            i++;
            continue;
        }
        bounce(i, i + 1);
    }
}


void AdvRivals::bounce(int i, int j) {

    /* the "pursuer" of 'i' is the civ with strictly fewer points than 'i'
       (pursuerIpts < iPts) that has the minimal point distance to 'i'.
       often that is 'i-1', but it could be a lower index if several civs
       have the same amount of configured points. one of the postconditions
       of 'randomizePoints' is that no civ overtake another in terms of
       start points, though ties can be broken. see GlobalDefines_advri.xml
       for the postconditions. */
    int iPts = sbp[i]->startPoints_configured();
    int pursuerI;
    int pursuerIpts;
    /* don't count to >=0 b/c that's the human player whose points aren't randomized
      (another postcondition) */
    for(pursuerI = i; pursuerI > 0; pursuerI--) {
        pursuerIpts = sbp[pursuerI]->startPoints_configured();
        if(pursuerIpts < iPts)
            break;
    }
    int deltaMaxNeg = round((iPts - pursuerIpts) / 2.0);
    int jPts = sbp[j]->startPoints_configured();
    int pursuedByJ;
    int pursuedByJpts;
    for(pursuedByJ = j; pursuedByJ < nCivs; pursuedByJ++) {
        pursuedByJpts = sbp[pursuedByJ]->startPoints_configured();
        if(pursuedByJpts > jPts)
            break;
    }
    int deltaMaxPos = round((pursuedByJpts - jPts) / 2.0);
    // to ensure that randomization leaves the sum of all points unchanged
    int deltaMax = deltaMaxNeg < deltaMaxPos ? deltaMaxNeg : deltaMaxPos;
    report += CvString::format("Bouncing apart civ ids %d (%d points) and %d (%d points)"
            " by at most plus/minus %d points\n",
            i, sbp[i]->startPoints_configured(),
            j, sbp[j]->startPoints_configured(),
            deltaMax);
    int delta = GC.getGameINLINE().getSorenRandNum(deltaMax,
            "(advri) start points randomization");
    sbp[i]->setStartPoints_actual(sbp[i]->startPoints_configured()
            - delta);
    sbp[j]->setStartPoints_actual(sbp[j]->startPoints_configured()
            + delta);
}


void AdvRivals::rearrangeStartingPlots() {

    report += "Assignment of starting plots:\n";
    report += CvString::format("Assigning starting plots for %d civs\n", nCivs);
    if(nCivs < 2)
        return;
    vector<CvPlot*> availStartingPlot(nCivs);
    for(int i = 0; i < nCivs; i++)
        availStartingPlot[i] = sbp[i]->startingPlot();
    vector<vector<int> > distances(nCivs, vector<int>(nCivs));
    int maxDist = -1;
    int anchor1 = -1,
        anchor2 = -1;
    for(int i = 0; i < nCivs; i++)
        for(int j = 0; j < nCivs; j++) {
            int d = dist(availStartingPlot[i], availStartingPlot[j], &availStartingPlot);
            distances[i][j] = d;
            if(d > maxDist) {
                maxDist = d;
                anchor1 = i;
                anchor2 = j;
            }
        }
    report += CvString::format("Maximal distance found to be %d\n", maxDist);
    report += CvString::format("Anchor plots: id %d and id %d\n", anchor1, anchor2);
    int civCount1 = countCivsInArea(availStartingPlot[anchor1]->getArea(), &availStartingPlot);
    int civCount2 = countCivsInArea(availStartingPlot[anchor2]->getArea(), &availStartingPlot);
    report += CvString::format("Counted %d civs on continent of anchor 1\n", civCount1);
    report += CvString::format("Counted %d civs on continent of anchor 2\n", civCount2);
    bool humanSecluded;
    prefSecluded = GC.getDefineINT("ADVRI_PREFER-SECLUDED-START");
    switch(prefSecluded) {
        case -1: humanSecluded = false; break;
        case  1: humanSecluded = true; break;
        default: humanSecluded = bernoulliSuccess(0.5);
    }
    int anchorHuman, anchorAi;
    if(humanSecluded ? civCount2 < civCount1 : civCount2 > civCount1) {
        anchorHuman = anchor2;
        anchorAi = anchor1;
    }
    else {
        anchorHuman = anchor1;
        anchorAi = anchor2;
    }
    vector<CvPlot*> designatedPlot(nCivs);
    designatedPlot[0] = availStartingPlot[anchorHuman];
    report += CvString::format("Human start: (%d,%d)\n", designatedPlot[0]->getX(),
        designatedPlot[0]->getY());
    designatedPlot[nCivs - 1] = availStartingPlot[anchorAi];
    report += CvString::format("Strongest AI start: (%d,%d)\n",
        designatedPlot[nCivs - 1]->getX(), designatedPlot[nCivs - 1]->getY());
    // available no more:
    availStartingPlot[anchorHuman] = availStartingPlot[anchorAi] = NULL;
    // v0.3: use "quartile" criterion (4) exclusively (subject to further changes)
    int criteriaSequence[6] = {4,4,4,4,4,4};//{0, 0, -1, -1, 1, 1};
    for(int i = 0; i < nCivs - 2; i++) { // two already assigned
        int nextCiv = -1;
        switch(criteriaSequence[i % 6]) {
        case 0: { // criterion "0": select the "middlest" civ
            int minDistFromMiddle = 3 * nCivs;
            int middlest = -1;
            for(int j = nCivs - 1; j >= 0; j--) {
                if(designatedPlot[j] != NULL)
                    continue;
                int dMiddle = abs(2 * j + 1 - nCivs);
                if(dMiddle < minDistFromMiddle) {
                    minDistFromMiddle = dMiddle;
                    middlest = j;
                }
            }
            nextCiv = middlest;
            break;
        }
        case -1: { // criterion -1: select lowest (lowliest) civ
            for(int j = 0; j < nCivs; j++)
                if(designatedPlot[j] == NULL) {
                    nextCiv = j;
                    break;
                }
            break;
        }
        case 1: { // criterion 1: select the highest civ
            for(int j = nCivs - 1; j >= 0; j--)
                if(designatedPlot[j] == NULL) {
                    nextCiv = j;
                    break;
                }
            break;
        }
        case 4: { // criterion 4: select the civ nearest to the quartile
            int minDistFromQuartile = 4 * nCivs;
            int closest = -1;
            for(int j = nCivs - 1; j >= 0; j--) {
                if(designatedPlot[j] != NULL)
                    continue;
                int dQuartile = abs(4 * j - nCivs);
                if(dQuartile < minDistFromQuartile) {
                    minDistFromQuartile = dQuartile;
                    closest = j;
                }
            }
            nextCiv = closest;
            break;
        }}
        report += CvString::format("Giving a plot to civ id %d now, based on criterion %d\n",
            nextCiv, criteriaSequence[i % 6]);
        int anchor = nCivs - 1;
        int rem = i % 4;
        if(rem == 3 || rem == 0) {
            report += "Assigning the remaining plot closest to human\n";
            anchor = 0;
        }
        else report += "Assigning the remaining plot closest to strongest AI\n";
        int minDistAnchor = INT_MAX;
        int nextPlot = -1;
        for(int j = 0; j < nCivs; j++) {
            if(availStartingPlot[j] == NULL)
                continue;
            CvPlot* anchorPlot = designatedPlot[anchor];
            int dAnchor = dist(anchorPlot, availStartingPlot[j], NULL);
            if(dAnchor < minDistAnchor) {
                minDistAnchor = dAnchor;
                nextPlot = j;
            }
        }
        designatedPlot[nextCiv] = availStartingPlot[nextPlot];
        report += CvString::format("Closest plot is (%d,%d) at distance %d\n",
            availStartingPlot[nextPlot]->getX(), availStartingPlot[nextPlot]->getY(),
            minDistAnchor);
        availStartingPlot[nextPlot] = NULL;
    }
    for(int i = 0; i < nCivs; i++)
        civs[i]->setStartingPlot(designatedPlot[i]);
    report += "Done reassigning starting plots\n";
}


int AdvRivals::dist(CvPlot* p, CvPlot* q, vector<CvPlot*>* startingPlot) {

    int r = GC.getMapINLINE().calculatePathDistance(p, q);
    if(r < 0) {
        r = stepDistance(p->getX(), p->getY(), q->getX(), q->getY()) + 10000;
        if(startingPlot != NULL) {
            /* the following hack leads to the anchors (player's starting plot
               and starting plot of the toughest AI) being placed on the continents
               with the highest number of starting plots, i.e. (practically) always
               avoids insular starts for the first anchor, and for the second unless
               there's only one continent with more than one civ): */
            r += 100 * (countCivsInArea(p->getArea(), startingPlot)
                      + countCivsInArea(q->getArea(), startingPlot));
        }
    }
    return r;
}


int AdvRivals::countCivsInArea(int area, vector<CvPlot*>* startingPlot) {

    int r = 0;
    for(unsigned int i = 0; i < startingPlot->size(); i++) {
        if(area == (*startingPlot)[i]->getArea())
            r++;
    }
    return r;
}


int AdvRivals::round(double d) {

    return (int)floor(0.5 + d);
}


bool AdvRivals::bernoulliSuccess(double pr) {

    int chancePerMill = round(pr * 1000.0);
    return GC.getGameINLINE().getSorenRandNum(1000, "AdvRivals::bernoulliSuccess")
            < chancePerMill;
}


// wrt. start points; 1st rank received the most
int AdvRivals::rank(int civId) {

    int r;
    for(r = nCivs; r > 0; r--)
        if(sbp[nCivs - r]->id() == civId)
            break;
    return r;
}


void AdvRivals::reset() {

    civs.clear();
    sbp.clear();
    pointsDisplayString = NULL;
}


void AdvRivals::updatePointsDisplayString() {

    if(sbp[nCivs - 1]->startPoints_configured() == 0) {
        pointsDisplayString = new wstring();
        return;
    }
    char const* ind = "\t"; // indent
    char const* doubleIndent = "\t\t";
    wostringstream s;
    s << "Advanced Start (" << speed->getDescription()
            << " speed)\n"<< ind << "Your points: ";
    FAssert(human != NULL);
    s << human->startPoints_configured();
    /*if(prefSecluded != 0)
        s << " (starting on the " << (prefSecluded > 0 ? "second " : "") <<
                "most populous continent)";*/ // not that important
    s << "\n" << ind << "Rivals': ";
    bool showActual = (GC.getDefineINT("ADVRI_SHOW-EFFECT-OF-RANDOMIZATION") > 0);
    for(int i = 0; i < nCivs; i++) {
        if(civs[i]->equals(human))
            continue;
        int pts = showActual ? civs[i]->startPoints_actual() :
                               civs[i]->startPoints_configured();
        s << pts << ", ";
        if((i + 1) % 4 == 0 && i + 1 < nCivs)
            s << "\n" << doubleIndent;
    }
    s.seekp(((long)s.tellp()) - 2); // drop final comma
    s << "\n" << ind;
    if(randPoints && !showActual)
        s << "(not shown: randomization)" << std::endl;
    pointsDisplayString = new wstring(s.str());
}


void AdvRivals::setBarbarianStartTurn(int gameStartTurn) {

    m_barbarianStartTurn = gameStartTurn
            + GC.getGameINLINE().turnsElapsedForBarbarianCreation();
}


void AdvRivals::setResearchPercent() {

    /* not a good formula.
       it's difficult to adjust research speed as there's an interplay of start era,
       start points, start turn and inflation.
       Advanced Starts seem to lead to rather many cities and citizens, and comparatively
       few technologies. Start eras later than Ancient tend to have the opposite effect
       -- many techs initially, but few cities.
       The current solution is a shot in the dark: Given era and start points, increase
       the start turn moderately (see computeStartTurn), don't touch inflation, and
       leave research mostly as configured in CIV4HandicapSInfo.xml, unless the player
       configures considerably more or fewer points than the standard setting. */
    m_researchPercent = 100;
    if(GC.getGameINLINE().getStartEra() == 0)
        // max plus mean should be a good indicator for the frontrunners' research speed
        m_researchPercent += round((maxStartPoints() + meanStartPoints()
                - scaleToGrowth(2000)) / 100.0);
    m_researchPercent = std::min(150, m_researchPercent);
}


int AdvRivals::getFoundingPrice() {

    /* price of a settler would be fine, but the DLL knows nothing of settlers
      (or any specific unit classes for that matter) */
    for(int i = 0; i < GC.getNumUnitClassInfos(); i++) {
        CvUnitInfo& u = GC.getUnitInfo((UnitTypes)GC.getUnitClassInfo((UnitClassTypes)i).
                getDefaultUnitIndex()); // each unit class has a default unit
        if(u.isFound()) // if that default unit can found cities, it'll do
            return u.getAdvancedStartCost();
    }
    FAssert(false);
    return -1;
}


void AdvRivals::sortByPoints() {

    sbp = vector<MajorCiv*>(civs);
    sbp[0] = human;
    std::sort(sbp.begin() + 1, sbp.end(), isLeftLessThanRight);
}


int AdvRivals::maxStartPoints() {

    return sbp[nCivs - 1]->startPoints_actual();
}


double AdvRivals::meanStartPoints() {

    double sum = 0;
    for(int i = 0; i < nCivs; i++)
        sum += sbp[i]->startPoints_actual();
    return sum / nCivs;
}


bool AdvRivals::isLeftLessThanRight(MajorCiv* left, MajorCiv* right) {

    return left->startPoints_configured() < right->startPoints_configured();
}




int MajorCiv::foundingPrice;

MajorCiv::MajorCiv(CvPlayer& p) : p(p), actualPointsDone(false) {}

void MajorCiv::setStartPoints_configured(int pts) {

    startPoints_c = correct(pts);
    actualPointsDone = false;
}

void MajorCiv::setStartPoints_actual(int pts) {

    startPoints_a = correct(pts);
    actualPointsDone = true;
}

void MajorCiv::assignStartPoints() {

    if(!actualPointsDone) {
        startPoints_a = startPoints_c;
        actualPointsDone = true;
    }
    p.setAdvancedStartPoints(startPoints_a);
}

int MajorCiv::startPoints_configured() {

    /* '-1' is the internal code for a normal/ non-advanced start,
    also what is passed to the CvPlayer object.
    the outside world should use '0 (or less)' for that.
    TODO Just confusing? 0 everywhere should actually
    be no problem. */
    return std::max(0, startPoints_c);
}

int MajorCiv::startPoints_actual() {

    return startPoints_a;
}

bool MajorCiv::equals(MajorCiv* other) {

    return p.getID() == other->p.getID();
}

CvPlot* MajorCiv::startingPlot() {

    return p.getStartingPlot();
}

void MajorCiv::setStartingPlot(CvPlot* plot) {

    p.setStartingPlot(plot, true);
}

int MajorCiv::era() {

    return p.getCurrentEra();
}

int MajorCiv::id() {

    return p.getID();
}

bool MajorCiv::isHuman() {

    return p.isHuman();
}

int MajorCiv::correct(int pts) {

    // > 0 but not enough to found a city doesn't make sense
    if(pts < foundingPrice)
        pts = isHuman() ? -1 : foundingPrice;
                // ^use '-1' to signal normal start (non-advanced)
    return pts;
    /* Unmodded behavior: the initial city costs all the points
       you have if you have between 1 and the actual cost of a city.
       Maybe also okay? In any case, the Settings screen shouldn't
       show 50 points if 100 can actually be spent.
       The requirement that AIs have to start Advanced could
       probably be dropped easily. */
}

void MajorCiv::setFoundingPrice(int c) {

    foundingPrice = c;
}